/*****************************************************************/
/*                                                               */
/* Produit de matrice par double diffusion                       */
/*                                                               */
/*****************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <mpi.h>
#include <math.h>
#include <pthread.h>
#include <pthread.h>
#include <semaphore.h>


/******* Fonctions d'affichage ***********/
#define VOIRD(expr) do {printf("P%d (%d,%d) <%s,%.3d> : \t{ " #expr " = %d }\n", \
                               my_id,i_row,i_col,__FUNCTION__,__LINE__,expr);fflush(NULL);} while(0)
/* #define VOIR_SEMVALUE(expr) do {int value; value = 0 ; \ */
/*                                sem_getvalue(expr,&value); \ */
/* 	                       printf("P%d (%d,%d) <%s,%.3d> : \t{ " #expr " = %d }\n", \ */
/*                                my_id,i_row,i_col,__FUNCTION__,__LINE__,value);fflush(NULL);} while(0) */

#define VOIR_SEMVALUE(expr) 
#define PRINT_MESSAGE(expr) do {printf("P%d (%d,%d) <%s,%.3d> : \t" #expr "\n", \
                                      my_id,i_row,i_col,__FUNCTION__,__LINE__);fflush(NULL);} while(0)

typedef struct { 
  int row;   /* le nombre de lignes   */
  int col;   /* le nombre de colonnes */
  double* data; /* les valeurs           */
} matrix_t;

typedef struct { 
  matrix_t A;
  matrix_t B;
  matrix_t C;
  matrix_t buf_col[2];
  matrix_t buf_row[2];
  sem_t *sem_perso;
  sem_t *sem_main;
} thread_arg_t;

int my_id, nb_proc;
int nb_col,nb_row;       /* Largeur et hauteur de la grille */
MPI_Comm MPI_HORIZONTAL; /* Communicateur pour diffusions horizontales */
MPI_Comm MPI_VERTICAL;   /* Communicateur pour diffusions verticales */
int i_col,i_row;         /* Position dans la grille */


/*******************************************/
/* initialisation alatoire de la matrice  */
/*******************************************/
void mat_init_alea(matrix_t *mat,int width, int height)
{
  int i;

  mat->row = height;
  mat->col = width;

  mat->data=(double*)calloc(height*width,sizeof(double));
  for(i=0 ; i<height*width ; i++)
    mat->data[i]=1.0*rand()/(RAND_MAX+1.0);
}

/*******************************************/
/* initialisation  0 de la matrice        */
/*******************************************/
void mat_init_empty(matrix_t *mat,int width, int height)
{
  mat->row = height;
  mat->col = width;
  mat->data=(double*)calloc((mat->row)*(mat->col),sizeof(double));
}

/*******************************************/
/* affichage de la matrice                 */
/*******************************************/
void mat_display(matrix_t A)
{
  int i,j,t=0;

  printf("      ");
  for(j=0;j<A.col;j++)
    printf("%.3d ",j);
  printf("\n");
  printf("    __");
  for(j=0;j<A.col;j++)
    printf("____");
  printf("_\n");

  for(i=0;i<A.row;i++)
    {
      printf("%.3d | ",i);
      for(j=0;j<A.col;j++)
	printf("%.3g ",A.data[t++]);
      printf("|\n");
    }
  printf("    --");
  for(j=0;j<A.col;j++)
    printf("----");
  printf("-\n");
}

/*******************************************/
/* C+=A*B                                  */
/*******************************************/
void mat_mult(matrix_t A, matrix_t B, matrix_t C)
{
  int i,j,k,M,N,K;
  double *_A,*_B,*_C;
 
  _A=A.data;
  _B=B.data;
  _C=C.data;

  M = C.row;
  N = C.col;
  K = A.col;

  if((M!=A.row) || (N!=B.col) || (K!=B.row)) {
    PRINT_MESSAGE("Attention, tailles incompatibles");
    VOIRD(A.row);VOIRD(A.col);VOIRD(B.row);
    VOIRD(C.col);VOIRD(C.row);VOIRD(C.col);
    exit(1);
  }

  for(i=0 ; i<M ; i++)
    for(j=0 ; j<N ; j++)
      for(k=0 ; k<K ; k++) 
	_C[i*N+j]+=_A[i*K+k]*_B[k*N+j];
}

/*******************************************/
/* Initialisations la grille               */
/*******************************************/
void init_communicateurs(void)
{
  int i;
  nb_col=nb_proc; nb_row=1;
  /* Amliorez la disposition */
  for(i=sqrt(nb_proc); i>=1 ; i--) 
    if (nb_proc%i==0) {
      nb_row=i; nb_col=nb_proc/nb_row; break;
    }
  i_col = my_id % nb_col;
  i_row = my_id / nb_col;

  /* Partitionnez MPI_COMM_WORLD en communicateurs pour les lignes 
     et en communicateurs pour les colonnes */
  MPI_Comm_split( MPI_COMM_WORLD, i_row, i_col, &MPI_HORIZONTAL ); 
  MPI_Comm_split( MPI_COMM_WORLD, i_col, i_row, &MPI_VERTICAL ); 
  
  /* Vrifiez que votre rang dans le nouveau communicateur 
     correspond bien  celui que vous attendez. */
  MPI_Comm_rank(MPI_HORIZONTAL,&i_col);
  MPI_Comm_rank(MPI_VERTICAL,&i_row);

}

/*******************************************/
/* Produit de matrice par double diffusion */
/*******************************************/
void diffuse(matrix_t A,matrix_t B,matrix_t C, int k,
	     matrix_t buf_col,matrix_t buf_row)
{
  if(k/A.col == i_col) {
    int i;
    for(i=0;i<C.row;i++) 
      buf_col.data[i]=A.data[i*A.col+(k%A.col)];
  }
  MPI_Bcast(buf_col.data, C.row, MPI_DOUBLE, k/A.col, MPI_HORIZONTAL);
  
  /* premire diffusion verticale */
  if(k/B.row == i_row) {
    int j;
    for(j=0;j<C.col;j++) 
      buf_row.data[j]=B.data[(k%B.row)*B.col+j];
  }
  MPI_Bcast(buf_row.data, C.col, MPI_DOUBLE, k/B.row, MPI_VERTICAL);
}

void communications(thread_arg_t *arg) {
  matrix_t buf_col[2], buf_row[2];
  matrix_t A, B, C;
  int k;  

  buf_col[0]=arg->buf_col[0];
  buf_col[1]=arg->buf_col[1];
  buf_row[0]=arg->buf_row[0];
  buf_row[1]=arg->buf_row[1];
  A=arg->A;
  B=arg->B;
  C=arg->C;

  for(k=1; k<(nb_col*C.col) ; k++) {
    /* double diffusion */
    PRINT_MESSAGE("On attend de pouvoir diffuser");
    sem_wait(arg->sem_perso);
    PRINT_MESSAGE("On diffuse");
    diffuse(A,B,C,k,buf_col[k%2],buf_row[k%2]);
    PRINT_MESSAGE("Communications termines: on libre le thread principal");
    sem_post(arg->sem_main);
  }
}

void computations(thread_arg_t *arg) {
  matrix_t buf_col[2], buf_row[2], C;
  int k;  
  buf_col[0]=arg->buf_col[0];
  buf_col[1]=arg->buf_col[1];
  buf_row[0]=arg->buf_row[0];
  buf_row[1]=arg->buf_row[1];
  C=arg->C;

  for(k=0; k<(nb_col*C.col)-1 ; k++) {
    /* produit de matrices en local */
    PRINT_MESSAGE("On attend de pouvoir calculer");
    sem_wait(arg->sem_perso);
    PRINT_MESSAGE("On diffuse");
    mat_mult(buf_col[k%2], buf_row[k%2],C);
    PRINT_MESSAGE("Calculs termins : on libre le thread principal");
    sem_post(arg->sem_main);  
  }
}

void parallel_mat_mult(matrix_t A, matrix_t B, matrix_t C)
{
  matrix_t buf_col[2], buf_row[2];
  thread_arg_t arg_comm,arg_comp;
  pthread_t communications_thread,computations_thread;
  sem_t sem_main,sem_comm,sem_comp;
  int k;
  
  /* Initialisation des buffers */
  mat_init_empty(&buf_col[0],1,C.row);
  mat_init_empty(&buf_row[0],C.col,1);
  mat_init_empty(&buf_col[1],1,C.row);
  mat_init_empty(&buf_row[1],C.col,1);
  
  /* Cration des arguments */
  arg_comp.A=arg_comm.A=A;
  arg_comp.B=arg_comm.B=B;
  arg_comp.C=arg_comm.C=C;
  arg_comp.buf_col[0]=arg_comm.buf_col[0]=buf_col[0];
  arg_comp.buf_col[1]=arg_comm.buf_col[1]=buf_col[1];
  arg_comp.buf_row[0]=arg_comm.buf_row[0]=buf_row[0];
  arg_comp.buf_row[1]=arg_comm.buf_row[1]=buf_row[1];
  arg_comp.sem_main=arg_comm.sem_main=&sem_main;
  arg_comp.sem_perso=&sem_comp;
  arg_comm.sem_perso=&sem_comm;

  PRINT_MESSAGE("Initialisation de la synchro");
  /* Initialisation de la synchro */
  sem_init (&sem_comp,0,0);
  VOIR_SEMVALUE(&sem_comp);
  sem_init (&sem_comm,0,0);
  VOIR_SEMVALUE(&sem_comm);
  sem_init (&sem_main,0,0);
  VOIR_SEMVALUE(&sem_main);


  /* Cration des deux threads */
  PRINT_MESSAGE("Cration du thread de communications");
  pthread_create(&communications_thread, NULL, communications, &arg_comp);
  PRINT_MESSAGE("Cration du thread de calcul");
  pthread_create(&computations_thread, NULL, computations, &arg_comm);

  k=0;
  PRINT_MESSAGE("Prmire diffusion");
  /* premire diffusion horizontale */
  diffuse(A,B,C,k,buf_col[k%2],buf_row[k%2]);

  /* Diffusion de k et calcul de k-1 en parallle */
  for(k=1; k<(nb_col*C.col) ; k++) {
    VOIRD(k);
    PRINT_MESSAGE("On libre les communications");
    sem_post(&sem_comm);
    PRINT_MESSAGE("On libre les calculs");
    sem_post(&sem_comp);
    PRINT_MESSAGE("On attends");
    sem_wait(&sem_main);
    sem_wait(&sem_main);
  }
  PRINT_MESSAGE("Dernier calcul");
  mat_mult(buf_col[k%2], buf_row[k%2],C);
}

int main(int argc, char **argv) 
{
  int taille=0;

  MPI_Init(&argc,&argv);
  MPI_Comm_size(MPI_COMM_WORLD,&nb_proc);
  MPI_Comm_rank(MPI_COMM_WORLD,&my_id);
  
  if(argc==0)
    fprintf(stderr,"Usage : %s <taille de matrice>",argv[0]);
  else 
    taille = atoi(argv[1]);

  init_communicateurs();
  
  {
    matrix_t A,B,C;
    
    mat_init_alea(&A,taille/nb_col,taille/nb_row);
    mat_init_alea(&B,taille/nb_col,taille/nb_row);
    mat_init_empty(&C,taille/nb_col,taille/nb_row);
    parallel_mat_mult(A,B,C);
    MPI_Barrier(MPI_COMM_WORLD);
    if(my_id==0) mat_display(C);
  }
  MPI_Finalize();
  
  return (0);
}
