/*****************************************************************/
/*                                                               */
/* Outer matrix product                                          */
/*                                                               */
/*****************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <mpi.h>
#include <math.h>
#include <pthread.h>
#include <semaphore.h>


/******* Display functions ***********/
#define VOIRD(expr) do {printf("P%d (%d,%d) <%s,%.3d> : \t{ " #expr " = %d }\n", \
                               my_id,i_row,i_col,__FUNCTION__,__LINE__,expr);fflush(NULL);} while(0)
/* #define VOIR_SEMVALUE(expr) do {int value; value = 0 ; \ */
/*                                sem_getvalue(expr,&value); \ */
/* 	                       printf("P%d (%d,%d) <%s,%.3d> : \t{ " #expr " = %d }\n", \ */
/*                                my_id,i_row,i_col,__FUNCTION__,__LINE__,value);fflush(NULL);} while(0) */

#define VOIR_SEMVALUE(expr) 
#define PRINT_MESSAGE(expr) do {printf("P%d (%d,%d) <%s,%.3d> : \t" #expr "\n", \
                                      my_id,i_row,i_col,__FUNCTION__,__LINE__);fflush(NULL);} while(0)

typedef struct { 
  int row;   /* number of lines       */
  int col;   /* number of columns     */
  double* data; /* data                  */
} matrix_t;

typedef struct { 
  matrix_t A;
  matrix_t B;
  matrix_t C;
  matrix_t buf_col[2];
  matrix_t buf_row[2];
  sem_t *sem_perso;
  sem_t *sem_main;
} thread_arg_t;

int my_id, nb_proc;
int nb_col,nb_row;       /* Width and heigh of the tore     */
MPI_Comm MPI_HORIZONTAL; /* Communicator for horizontal broadcasts */
MPI_Comm MPI_VERTICAL;   /* Communicator for verticales boradcasts */
int i_col,i_row;         /* Position in the tore */


/*******************************************/
/* random matrix initialisation            */
/*******************************************/
void mat_init_alea(matrix_t *mat,int width, int height)
{
  int i;

  mat->row = height;
  mat->col = width;

  mat->data=(double*)calloc(height*width,sizeof(double));
  for(i=0 ; i<height*width ; i++)
    mat->data[i]=1.0*rand()/(RAND_MAX+1.0);
}

/*******************************************/
/* 0 initialiation matrix                  */
/*******************************************/
void mat_init_empty(matrix_t *mat,int width, int height)
{
  mat->row = height;
  mat->col = width;
  mat->data=(double*)calloc((mat->row)*(mat->col),sizeof(double));
}

/*******************************************/
/* matrix display                          */
/*******************************************/
void mat_display(matrix_t A)
{
  int i,j,t=0;

  printf("      ");
  for(j=0;j<A.col;j++)
    printf("%.3d ",j);
  printf("\n");
  printf("    __");
  for(j=0;j<A.col;j++)
    printf("____");
  printf("_\n");

  for(i=0;i<A.row;i++)
    {
      printf("%.3d | ",i);
      for(j=0;j<A.col;j++)
	printf("%.3g ",A.data[t++]);
      printf("|\n");
    }
  printf("    --");
  for(j=0;j<A.col;j++)
    printf("----");
  printf("-\n");
}

/*******************************************/
/* C+=A*B                                  */
/*******************************************/
void mat_mult(matrix_t A, matrix_t B, matrix_t C)
{
  int i,j,k,M,N,K;
  double *_A,*_B,*_C;
 
  _A=A.data;
  _B=B.data;
  _C=C.data;

  M = C.row;
  N = C.col;
  K = A.col;

  if((M!=A.row) || (N!=B.col) || (K!=B.row)) {
    PRINT_MESSAGE("Warning: incompatible sizes");
    VOIRD(A.row);VOIRD(A.col);VOIRD(B.row);
    VOIRD(C.col);VOIRD(C.row);VOIRD(C.col);
    exit(1);
  }

  for(i=0 ; i<M ; i++)
    for(j=0 ; j<N ; j++)
      for(k=0 ; k<K ; k++) 
	_C[i*N+j]+=_A[i*K+k]*_B[k*N+j];
}

/*******************************************/
/* Tore initialisation                     */
/*******************************************/
void init_communicateurs(void)
{
  nb_col=nb_proc; nb_row=1;
  /* Compute better #col/#row */

  /* Partition MPI_COMM_WORLD into communicators for lines 
     and communicators for columns */

  /* Check your rank in the new communicator is the good
     one. */ 
}

/*******************************************/
/* Outer matrix product                    */
/*******************************************/
void diffuse(matrix_t A,matrix_t B,matrix_t C, int k,
	     matrix_t buf_col,matrix_t buf_row)
{
    /* horizontal diffusion */
    /* vertical diffusion */
}

void communications(thread_arg_t *arg) {
  matrix_t buf_col[2], buf_row[2];
  matrix_t A, B, C;
  int k;  

  for(k=1; k<(nb_col*C.col) ; k++) {
    /* We wait to be able to broadcast */
    /* We broadcast */
    /* We allow parallel_mat_mult to run */
  }
}

void computations(thread_arg_t *arg) {
  matrix_t buf_col[2], buf_row[2], C;
  int k;  

  for(k=0; k<(nb_col*C.col)-1 ; k++) {
    /* We wait to be able to compute */
    /* We compute */
    /* We allow parallel_mat_mult to run */
  }
}

void parallel_mat_mult(matrix_t A, matrix_t B, matrix_t C)
{
  matrix_t buf_col[2], buf_row[2];
  thread_arg_t arg_comm,arg_comp;
  pthread_t communications_thread,computations_thread;
  sem_t sem_main,sem_comm,sem_comp;
  int k;
  
  /* Initialisation des buffers */
  mat_init_empty(&buf_col[0],1,C.row);
  mat_init_empty(&buf_row[0],C.col,1);
  mat_init_empty(&buf_col[1],1,C.row);
  mat_init_empty(&buf_row[1],C.col,1);
  
  /* Création des arguments des threads */

  /* Initialisation des sémaphores */

  /* Création des deux threads */

  /* première diffusion horizontale */
  k=0;

  /* Diffusion de k et calcul de k-1 en parallèle */
  for(k=1; k<(nb_col*C.col) ; k++) {
    /* On libère les communications */
    /* On libère les calculs */
    /* On attends */
  }
  /* Dernier calcul */
}

int main(int argc, char **argv) 
{
  int size=0;

  MPI_Init(&argc,&argv);
  MPI_Comm_size(MPI_COMM_WORLD,&nb_proc);
  MPI_Comm_rank(MPI_COMM_WORLD,&my_id);
  
  if(argc==0)
    fprintf(stderr,"Usage : %s <matrix size>",argv[0]);
  else 
    size = atoi(argv[1]);

  init_communicateurs();
  
  {
    matrix_t A,B,C;
    
    mat_init_alea(&A,size/nb_col,size/nb_row);
    mat_init_alea(&B,size/nb_col,size/nb_row);
    mat_init_empty(&C,size/nb_col,size/nb_row);
    parallel_mat_mult(A,B,C);
    MPI_Barrier(MPI_COMM_WORLD);
    if(my_id==0) mat_display(C);
  }
  MPI_Finalize();
  
  return (0);
}
